/*! 
* This Source Code Form is subject to the terms of the Mozilla Public
* License, v. 2.0. If a copy of the MPL was not distributed with this file,
* You can obtain one at http://mozilla.org/MPL/2.0/.
*
* Copyright(c) 2009 Apogee Instruments, Inc. 
* \class IEthernet 
* \brief Interface class for the ethernet 
* 
*/ 


#ifndef IETHERNET_INCLUDE_H__ 
#define IETHERNET_INCLUDE_H__ 

#include <vector>
#include <string>
#include "CameraStatusRegs.h" 

class IEthernet 
{ 
    public: 
        virtual ~IEthernet(); 

        /*!
         *  Reads camera control registers
         * \param[in] FpgaReg Register to read.
         */
        virtual uint16_t ReadReg(uint16_t FpgaReg ) = 0;

        /*!
         *  Writes data to camera control registers
         * \param[in] FpgaReg Register to write to.
         * \param[out] FpgaData Data value in the register
         */
        virtual void WriteReg(uint16_t FpgaReg, 
            uint16_t FpgaData ) = 0;

        /*!
         *  Moves the data from the camera to the host PC.  Called at different
         *  times depending on the cameras acquistion mode
         * \param[out] ImageData Image from the camera
         * \param[in] NumBytesExpected Number of bytes that should move from the camera to the PC
         */
        virtual void GetImageData(std::vector<uint16_t> & ImageData, 
					       const uint32_t NumBytesExpected) = 0;	

        /*!
         * Prime the IO interface for an upcoming image transfer.
         * \param[in] Rows Image height in pixels
         * \param[in] Cols Image width in pixels
         * \param[in] NumOfImages Number of images in the set
         */
        virtual void SetupImgXfer(uint16_t Rows, 
            uint16_t Cols, uint16_t NumOfImages, bool IsBulkSeq) = 0;

        /*!
         *  Cancel the pending image transfer.
         */
         virtual void CancelImgXfer() = 0;

        /*!
         *  Fetches the camera's status. 
         * \return The camera's status as packed data structure see
         *         CameraStatusRegs::AdvStatus and CameraStatusRegs::BasicStatus
         *         For the byte layout
         */
        virtual void GetStatus(CameraStatusRegs::AdvStatus & status) = 0;

        virtual void GetStatus(CameraStatusRegs::BasicStatus & status) = 0;

        virtual void GetMacAddress( std::string & Mac ) = 0;

        virtual void WriteSRMD( uint16_t reg, const std::vector<uint16_t> & data ) = 0;

        virtual std::string GetNetworkSettings() = 0;

        virtual std::string GetDriverVersion() = 0;
}; 

#endif