/* 
 * Apogee Alta CCD driver.
 * Copyright (C) 2012 The Random Factory
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include <iostream>
#include <string>
#include <vector>
#include <sstream>
#include <stdint.h>
#include <stdexcept>
#include <sys/time.h>


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "fitsio.h"
#include <math.h>
#include <unistd.h>
#include <time.h>

#include "camd.h"
#ifdef APOGEE_ALTA
#include "Alta.h"
#endif
#ifdef APOGEE_ASCENT
#include "Ascent.h"
#endif
#include "FindDeviceUsb.h"
#include "CameraInfo.h"
#include "apgSampleCmn.h"

namespace rts2camd
{

/**
 * Driver for Apogee Alta cameras.
 *
 * Based on test_alta.cpp from Alta source.
 *
 * @author Dave Mills <rfactory@theriver.com>
 */
class RTSAlta:public Camera
{
	public:
		RTSAlta (int argc, char **argv);
		virtual ~ RTSAlta (void);
		virtual int processOption (int in_opt);
		virtual int initHardware ();

		virtual int info ();

		virtual int setCoolTemp (float new_temp);
		virtual void afterNight ();

	protected:
		virtual int initChips ();
		virtual void initBinnings ();
		virtual int setBinning (int in_vert, int in_hori);
		virtual int startExposure ();
		virtual long isExposing ();
		virtual int stopExposure ();
		virtual int doReadout ();

		virtual int setValue (rts2core::Value * old_value, rts2core::Value * new_value);

		virtual size_t suggestBufferSize ()
		{
			return Camera::suggestBufferSize () + 4;
		}

	private:
#ifdef APOGEE_ALTA
		Alta * altaCamera;
#endif
#ifdef APOGEE_ASCENT
		Ascent * altaCamera;
#endif
		rts2core::ValueSelection * bitDepth;
		rts2core::ValueSelection *fanMode;

		rts2core::ValueSelection *coolerStatus;
		rts2core::ValueBool *coolerEnabled;

		rts2core::ValueInteger *cameraId;

		void setBitDepth (int newBit);
};

};

using namespace rts2camd;

///////////////////////////
// MAKE   TOKENS
std::vector<std::string> makeTokens(const std::string &str, const std::string &separator)
{
        std::vector<std::string> returnVector;
        std::string::size_type start = 0;
        std::string::size_type end = 0;

        while( (end = str.find(separator, start)) != std::string::npos)
        {
                returnVector.push_back (str.substr (start, end-start));
                start = end + separator.size();
        }

        returnVector.push_back( str.substr(start) );

        return returnVector;
}


////////////////////////////
//      GET             ADDRESS
std::string GetAddress( const std::string & msg )
{

        //search the find string for an attached camera address
        std::vector<std::string> params = makeTokens( msg, "," );
        std::vector<std::string>::iterator iter;

        for(iter = params.begin(); iter != params.end(); ++iter)
        {
           if( std::string::npos != (*iter).find("address=") )
           {
                 std::string DeviceAddr = makeTokens( (*iter), "=" ).at(1);
                 //only looking for one camera
                 return DeviceAddr ;
           }
        } //for

        std::string noOp;
        return noOp;
}


void RTSAlta::setBitDepth (int newBit)
{
	switch (newBit)
	{
		case 0:
			altaCamera->SetCcdAdcResolution (Apg::Resolution_SixteenBit);
		case 1:
			altaCamera->SetCcdAdcResolution (Apg::Resolution_TwelveBit);
	}
}

int RTSAlta::initChips ()
{
	setSize (altaCamera->GetRoiNumCols (), altaCamera->GetRoiNumRows (), 0, 0);
	pixelX = nan ("f");
	pixelY = nan ("f");
	//  gain = alta->m_ReportedGainSixteenBit;

	return Camera::initChips ();
}

void RTSAlta::initBinnings ()
{
	addBinning2D (1,1);
	addBinning2D (2,2);
	addBinning2D (3,3);
	addBinning2D (4,4);
	addBinning2D (8,8);
	addBinning2D (16,16);
	addBinning2D (32,32);
	addBinning2D (64,64);
	addBinning2D (3,3);
	addBinning2D (5,5);
	addBinning2D (6,6);
	addBinning2D (7,7);
	addBinning2D (10,10);
	addBinning2D (20,20);

        setBinning (1, 1);
}

int RTSAlta::setBinning (int in_vert, int in_hori)
{
	altaCamera->SetRoiBinCol (in_hori);
	altaCamera->SetRoiBinRow (in_vert);
	return Camera::setBinning (in_vert, in_hori);
}

int RTSAlta::startExposure ()
{

	// set region of intereset..
	altaCamera->SetRoiStartRow (chipUsedReadout->getYInt ());
	altaCamera->SetRoiStartCol (chipUsedReadout->getXInt ());
	altaCamera->SetRoiNumRows (chipUsedReadout->getHeightInt () / binningVertical ());
	altaCamera->SetRoiNumCols (chipUsedReadout->getWidthInt () / binningHorizontal ());

	altaCamera->StartExposure (getExposure (), getExpType () ? 0 : 1);

	return 0;
}

long RTSAlta::isExposing ()
{
	long ret;
	Apg::Status status;
	ret = Camera::isExposing ();
	if (ret > 0)
		return ret;

	status = altaCamera->GetImagingStatus ();

	if (status < 0)
	{
		logStream (MESSAGE_ERROR) << "During check for ImagingStatus, return < 0: " << status << sendLog;
		return -2;
	}
	if (status != Apg::Status_ImageReady)
		return 200;
	// exposure has ended..
	return -2;
}

int RTSAlta::stopExposure ()
{
	altaCamera->StopExposure (false);
	return Camera::stopExposure ();
}

int RTSAlta::doReadout ()
{
	int ret;
        std::vector<uint16_t> pImageData;

	altaCamera->GetImage (pImageData);
        copy(pImageData.begin(), pImageData.end(), getDataBuffer(0));

	ret = sendReadoutData (getDataBuffer (0), getWriteBinaryDataSize ());
	if (ret < 0)
		return -1;
	return -2;
}

int RTSAlta::setValue (rts2core::Value * old_value, rts2core::Value * new_value)
{
	if (old_value == bitDepth)
	{
		setBitDepth (new_value->getValueInteger ());
		return 0;
	}
	if (old_value == coolerEnabled)
	{
		altaCamera->SetCooler (((rts2core::ValueBool *)new_value)->getValueBool ());
		return 0;
	}
	if (old_value == fanMode)
	{
		altaCamera->SetFanMode ((Apg::FanMode)new_value->getValueInteger ());
		return 0;
	}

	return Camera::setValue (old_value, new_value);
}

RTSAlta::RTSAlta (int in_argc, char **in_argv):Camera (in_argc, in_argv)
{
  	createValue (cameraId, "camera_id", "camera ID", false);
	cameraId->setValueInteger (1);

	createTempAir ();
	createTempCCD ();
	createTempSet ();

	createExpType ();

	createValue (coolerStatus, "COOL_STA", "cooler status", true);
	coolerStatus->addSelVal ("OFF");
	coolerStatus->addSelVal ("RAMPING_TO_SETPOINT");
	coolerStatus->addSelVal ("CORRECTING");
	coolerStatus->addSelVal ("RAMPING_TO_AMBIENT");
	coolerStatus->addSelVal ("AT_AMBIENT");
	coolerStatus->addSelVal ("AT_MAX");
	coolerStatus->addSelVal ("AT_MIN");
	coolerStatus->addSelVal ("AT_SETPOINT");

	createValue (coolerEnabled, "COOL_ENA", "if cooler is enabled", true, RTS2_VALUE_WRITABLE);

	createValue (bitDepth, "BITDEPTH", "bit depth", true, RTS2_VALUE_WRITABLE, CAM_WORKING);
	bitDepth->addSelVal ("16 bit");
	bitDepth->addSelVal ("12 bit");

	createValue (fanMode, "FAN", "fan mode", true, RTS2_VALUE_WRITABLE);
	fanMode->addSelVal ("OFF");
	fanMode->addSelVal ("LOW");
	fanMode->addSelVal ("MEDIUM");
	fanMode->addSelVal ("HIGH");

	altaCamera = NULL;
	addOption ('b', NULL, 0, "switch to 12 bit readout mode; see alta specs for details");
	addOption ('c', NULL, 1, "night cooling temperature");
	addOption ('n', NULL, 1, "camera number (default to 1)");
}

RTSAlta::~RTSAlta (void)
{
	if (altaCamera)
	{
		altaCamera->CloseConnection ();
		delete altaCamera;
	}
}

int RTSAlta::processOption (int in_opt)
{
	switch (in_opt)
	{
		case 'b':
			bitDepth->setValueInteger (1);
			break;
		case 'c':
			nightCoolTemp->setValueCharArr (optarg);
			break;
		case 'n':
			cameraId->setValueCharArr (optarg);
			break;
		default:
			return Camera::processOption (in_opt);
	}
	return 0;
}

int RTSAlta::initHardware ()
{
        std::string ioInterface("usb");
        FindDeviceUsb lookUsb;
	
        std::string msg = lookUsb.Find();                
        std::string addr = apgSampleCmn::GetUsbAddress( msg );
        uint16_t id = apgSampleCmn::GetID( msg );
        uint16_t frmwrRev = apgSampleCmn::GetFrmwrRev( msg );
	
#ifdef APOGEE_ALTA
        altaCamera = (Alta *) new Alta();
        altaCamera->OpenConnection(ioInterface, addr, frmwrRev, id);
#endif

#ifdef APOGEE_ASCENT
        altaCamera = (Ascent *) new Ascent();
        altaCamera->OpenConnection(ioInterface, addr, frmwrRev, id);
#endif

	altaCamera->Init();

	altaCamera->Reset();
	altaCamera->CloseConnection ();
	sleep (2);
        altaCamera->OpenConnection(ioInterface, addr, frmwrRev, id);
	altaCamera->Init();

	// Do a system reset to ensure known state, flushing enabled etc
	altaCamera->Reset();

	// set data bits..

#ifdef APOGEE_ALTA
	strcpy (ccdType, "Alta ");
#endif
#ifdef APOGEE_ASCENT
	strcpy (ccdType, "Ascent ");
#endif
//	strncat (ccdType, altaCamera->GetSensor(), 10);
	sprintf (serialNumber, "%i", id);

	return initChips ();
}

int RTSAlta::info ()
{
	coolerStatus->setValueInteger (altaCamera->GetCoolerStatus ());
	coolerEnabled->setValueBool (altaCamera->IsCoolerOn ());
	tempSet->setValueDouble (altaCamera->GetCoolerSetPoint ());
	tempCCD->setValueFloat (altaCamera->GetTempCcd ());
	tempAir->setValueFloat (altaCamera->GetTempHeatsink ());
	fanMode->setValueInteger (altaCamera->GetFanMode ());
	return Camera::info ();
}

int RTSAlta::setCoolTemp (float new_temp)
{
	altaCamera->SetCooler(true);
	altaCamera->SetFanMode (Apg::FanMode_High);
	altaCamera->SetCoolerSetPoint (new_temp);
	return Camera::setCoolTemp (new_temp);
}

void RTSAlta::afterNight ()
{
	altaCamera->SetCooler (false);
	altaCamera->SetFanMode (Apg::FanMode_Low);
	altaCamera->SetCoolerSetPoint (100);
}

int main (int argc, char **argv)
{
	RTSAlta device (argc, argv);
	return device.run ();
}
	
	
