/**
 * Apogee filter wheel (USB based) driver.
 * Copyright (C) 2011 Petr Kubanek <petr@kubanek.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 

#include "filterd.h"
#include <iostream>
#include <string>
#include <vector>
#include <sstream>
#include <stdint.h>
#include <stdexcept>

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <time.h>
#ifdef APOGEE_ALTA
#include "ApogeeFilterWheel.h"
#endif
#ifdef APOGEE_ASCENT
#include "Ascent.h"
#endif
#include "FindDeviceUsb.h"
#include "CameraInfo.h"
#include "apgSampleCmn.h"


namespace rts2filterd
{

class RTSAlta:public Filterd
{
	public:
		RTSAlta (int in_argc, char **in_argv);
		virtual ~ RTSAlta (void);
	protected:
		virtual int processOption (int in_opt);
		virtual int init (void);

		virtual int getFilterNum (void);
		virtual int setFilterNum (int new_filter);

		virtual int homeFilter ();

	private:

		int filter_count;
		#ifdef APOGEE_ALTA
		ApogeeFilterWheel::Type filterType;
		ApogeeFilterWheel *dev;
		#else
		Ascent::FilterWheelType filterType;
		AscentFilterWheel *dev;
		#endif
};

}

using namespace rts2filterd;

RTSAlta::RTSAlta (int in_argc, char **in_argv):Filterd (in_argc, in_argv)
{
  	dev = NULL;
	#ifdef APOGEE_ALTA
	filterType = ApogeeFilterWheel::FW50_7S;
	#else
	filterType = Ascent::FW_UNKNOWN_TYPE;
	#endif

	addOption ('t', NULL, 1, "APN filter type (see Apogee.h for list; 0-5)");
}

RTSAlta::~RTSAlta (void)
{
	if (dev)
	{
		#ifdef APOGEE_ALTA
		dev->Close();
		#else
		dev->FilterWheelClose();
		#endif
		delete dev;
	}
}

int RTSAlta::processOption (int in_opt)
{
	switch (in_opt)
	{
		case 't':
			filterType = (ApogeeFilterWheel::Type) atoi (optarg);
			break;
		default:
			return Filterd::processOption (in_opt);
	}
	return 0;
}

int RTSAlta::init (void)
{
	int ret;
	ret = Filterd::init ();
	if (ret)
		return ret;
	
	
#ifdef APOGEE_ALTA
        FindDeviceUsb look4FilterWheel;
        std::string msg = look4FilterWheel.Find();
        if( !apgSampleCmn::IsDeviceFilterWheel( msg ) )
                {
                        logStream (MESSAGE_ERROR) <<"Device not a filter wheel = " + msg << sendLog;
                        return -1;
                }

        std::cout << "USB filter wheel found" << std::endl;
        std::string addr = apgSampleCmn::GetUsbAddress( msg );
        dev = new ApogeeFilterWheel();
        dev->Init(filterType,"1");
	ret = dev->GetMaxPositions ();
#endif

#ifdef APOGEE_ASCENT
        std::string ioInterface("usb");
        FindDeviceUsb look4cam;
        std::string msg = look4cam.Find();
        std::cout << msg << std::endl;
                
        std::string addr = apgSampleCmn::GetUsbAddress( msg );
        uint16_t id = apgSampleCmn::GetID( msg );
        uint16_t frmwrRev = apgSampleCmn::GetFrmwrRev( msg );
        
        dev = new Ascent();
        dev->OpenConnection( ioInterface, addr, frmwrRev, id );
        dev->FilterWheelOpen(filterType);
        ret = dev->GetFilterWheelMaxPositions();
#endif


	if (!ret)
	{
		logStream (MESSAGE_ERROR) << "cannot retrieve maximal numer of filter wheel positions" << sendLog;
		return -1;
	}
	return 0;
}

int RTSAlta::getFilterNum (void)
{
	unsigned long f;
#ifdef APOGEE_ALTA
	f = dev->GetPosition ();
#endif


#ifdef APOGEE_ASCENT
	f = dev->GetFilterWheelPos ();
#endif
	f--;
	if (f < 0)
		return -1;
	return (int) f;
}

int RTSAlta::setFilterNum (int new_filter)
{
	if (new_filter < -1 || new_filter >= filter_count)
		return -1;

#ifdef APOGEE_ALTA
	dev->SetPosition (new_filter+1);
#endif
	
#ifdef APOGEE_ASCENT
	dev->SetFilterWheelPos (new_filter+1);
#endif
	
	return Filterd::setFilterNum (new_filter);
}

int RTSAlta::homeFilter ()
{
	return setFilterNum (1);
}

int main (int argc, char **argv)
{
	RTSAlta device = RTSAlta (argc, argv);
	return device.run ();
}
